;*** GLONASS 1-bit DSP receiver INT4 routine - 02/10/1994 ***

;DSP common variables #offset(A0)
;$0000 long: 1ms interrupt counter
;$0004 word: interrupt delay reference count
;$0006 word: bits 0,4 count
;$0008 word: bits 1,5 count
;$000A word: bits 2,6 count
;$000C word: bits 3,7 count
;$000E long: magnitude sum average
;$0012 long: magnitude difference average
;$0016 word: averaging flag
;$0018 word: channel sequence ctr
;$001A word: data channel #
;$1040 320*word: sine table

;DSP channel variables: offset+$0200*(chNo-1)(A0)
;$001E word: channel PLL modulo
;$0020 long: previous sample 1ms epoch
;$0024 word: phase offset
;$0026 word: carrier phase
;$0028 long: carrier frequency
;$002C word: carrier-sync DCD
;$002E long: code offset
;$0032 word: code rate
;$0034 word: CTR phase increment
;$0036 word: code-sync DCD
;$0038 word: reference rate
;$003A 8*word: data buffer
;$004A word: old/new offset
;$004C long: reference 1ms epoch
;$0050 word: bit-sync DCD
;$0052 long: data accumulator

;$0056 word: signal level - treshold average
;$0058 long: carrier frequency average
;$005C word: 1st sample code offset reference
;$005E long: code offset average
;$0062 word: start 1ms epoch
;$0064 word: averaging counter /16
;$0066 word: buffer pointer
;$0068 16*4*long: buffer

;0178H long: manchester decoder>
;017CH long: differential decoder>
;0180H 11*word: last 176 bits>
;0196H word: counter /200>
;0198H word: data buffer pointer>
;019AH 8*(6*word+long): data buffer>

;*** start of INT4 routine ***

        ORG $00000000
        MOVE.B #$DE,(#common) ;common multiple latch all ctrs
        CLR.L (#intren)       ;reset & enable INT4
        MOVEM.L #$FFE0,-(A7)  ;D0-7,A0-2 to stack
        MOVEA.L #$FFFFFFFF,A0 ;set A0
        MOVEA.L A0,A2         ;set A2
        ADDQ.L #1,(A0)+       ;increment 1ms interrupt counter

;*** counter & control addresses ***

ic1ctr0 EQU $000E8081
ic1ctr1 EQU $000E8083
ic1ctr2 EQU $000E8089
ic2ctr0 EQU $000E8041
ic2ctr1 EQU $000E8043
ic2ctr2 EQU $000E8049
common  EQU $000E800B
intren  EQU $000E0082
moddiv  EQU 4500              ; 1ms divider modulo
modulo  EQU $11940000         ; 4500*65536
rbit1   EQU $000E0041
sbit1   EQU $000E0043
rbit0   EQU $000E0049
sbit0   EQU $000E004B
maxcar  EQU 1024
maxcod  EQU 1024
maxbit  EQU 256
rcbit   EQU $000E00C9         ; synthesizer modulo serial intf
scbit   EQU $000E00CB
rdbit   EQU $000E00C1
sdbit   EQU $000E00C3
rsbit   EQU $000E0089
ssbit   EQU $000E008B

;read counters & compute differences

        MOVEQ #0,D0
        MOVE.B (#ic1ctr1),D1  ; ic1ctr1 = 1ms divider
        MOVE.B (#ic1ctr1),D0
        ASL.L #8,D0
        MOVE.B D1,D0
        MOVE.W (A0),D1        ;interrupt delay reference count
        MOVE.W D0,(A0)+
        MOVE.W #moddiv,D2     ;reference D2.W
        ADD.W D1,D2
        SUB.W D0,D2
        MOVE.B (#ic2ctr0),D1  ;bits 0,4
        MOVE.B (#ic2ctr0),D0
        ASL.L #8,D0
        MOVE.B D1,D0
        MOVE.W (A0),D4        ;bits 0,4 difference D4.W
        MOVE.W D0,(A0)+
        SUB.W D0,D4
        ADD.W D4,D4
        SUB.W D2,D4
        MOVE.B (#ic2ctr1),D1  ;bits 1,5
        MOVE.B (#ic2ctr1),D0
        ASL.L #8,D0
        MOVE.B D1,D0
        MOVE.W (A0),D5        ;bits 1,5 difference D5.W
        MOVE.W D0,(A0)+
        SUB.W D0,D5
        ADD.W D5,D5
        SUB.W D2,D5
        MOVE.B (#ic1ctr2),D1  ;bits 2,6
        MOVE.B (#ic1ctr2),D0
        ASL.L #8,D0
        MOVE.B D1,D0
        MOVE.W (A0),D6        ;bits 2,6 difference D6.W
        MOVE.W D0,(A0)+
        SUB.W D0,D6
        ADD.W D6,D6
        SUB.W D2,D6
        MOVE.B (#ic2ctr2),D1  ;bits 3,7
        MOVE.B (#ic2ctr2),D0
        ASL.L #8,D0
        MOVE.B D1,D0
        MOVE.W (A0),D7        ;bits 3,7 difference D7.W
        MOVE.W D0,(A0)+
        SUB.W D0,D7
        ADD.W D7,D7
        SUB.W D2,D7

;compute A,D magnitude & I,Q components

        MOVE.W D4,D0          ;squared early magnitude
        MULS D0,D0
        MOVE.W D5,D2
        MULS D2,D2
        ADD.L D2,D0
        MOVE.W D4,D2
        BPL.B abs1
        NEG.W D2
abs1    MOVE.W D5,D3
        BPL.B abs2
        NEG.W D3
abs2    CMP.W D3,D2           ;square root approx.
        BCS.B sqrt1
        ASR.W #1,D3
        BRA.B sqrt2
sqrt1   ASR.W #1,D2
sqrt2   ADD.W D3,D2
        BNE.B sqrt3
        MOVEQ #1,D2
sqrt3   DIVU D2,D0            ;early magnitude D0.L
        ADD.W D0,D2
        MOVEQ #0,D0
        MOVE.W D2,D0

        MOVE.W D6,D1          ;squared late magnitude
        MULS D1,D1
        MOVE.W D7,D2
        MULS D2,D2
        ADD.L D2,D1
        MOVE.W D6,D2
        BPL.B abs3
        NEG.W D2
abs3    MOVE.W D7,D3
        BPL.B abs4
        NEG.W D3
abs4    CMP.W D3,D2           ;square root approx.
        BCS.B sqrt4
        ASR.W #1,D3
        BRA.B sqrt5
sqrt4   ASR.W #1,D2
sqrt5   ADD.W D3,D2
        BNE.B sqrt6
        MOVEQ #1,D2
sqrt6   DIVU D2,D1            ;late magnitude D1.L
        ADD.W D1,D2
        MOVEQ #0,D1
        MOVE.W D2,D1
        ADD.W D6,D4           ;I to D4.W
        ADD.W D7,D5           ;Q to D5.W
        MOVE.L (A0),D2        ;sum accumulation
        ADD.L D0,D2
        ADD.L D1,D2
        MOVE.L D2,(A0)+
        MOVE.L (A0),D2        ;difference accumulation
        ADD.L D0,D2
        SUB.L D1,D2
        MOVE.L D2,(A0)+

;valid data flag & return fm interrupt

        BSR.B vdflag
        MOVEM.L (A7)+,#$07FF  ;restore D0-7,A0-2
        RTE                   ;return
vdflag  MOVE.W (A0),D0        ;invalid data?
        BPL.B validd
        BSR.W srx
        BSR.W stx
        RTS
validd  ADDQ.W #1,D0          ;increment valid flag
        MOVE.W D0,(A0)+

        SUBQ.W #2,D0          ;reset magnitude accumulators?
        BPL.B count6
        CLR.L #$000E(A2)
        CLR.L #$0012(A2)
        BSR.W srx
        BSR.W stx
        RTS

;count 6 - set next channel parms

count6  CMPI.W #6,D0
        BNE.W chadd
        MOVE.W (A0),D1        ;get next channel #
        ADDQ.W #1,D1
        MOVE.W #$0002(A0),D2
        MOVE.W D1,D3
        ANDI.W #1,D3
        BNE.B ndatach
        EXT.L D1
        DIVU #3,D1
        EXT.L D1
        DIVU #3,D1
        SWAP D1
        ADDQ.W #1,D1
        ADD.W D1,D2
ndatach MOVEA.L A0,A1         ;set ch address A1
        ADDQ.L #8,A1
        ANDI.L #3,D2
        LSL.L #8,D2
        LSL.L #1,D2
        ADDA.L D2,A1

        MOVE.L (A1),D2        ;delta epoch to D3.L
        MOVE.L (A2),D3
        MOVE.L D3,(A1)+
        SUB.L D2,D3
        MOVE.L #$000A(A1),D0  ;old offset & rate
        MOVE.W #$000E(A1),D1
        MULS D3,D1            ;delta_epoch*rate+old_offset
        ASR.L #1,D1
        ADD.L D1,D0
        BPL.B pluoff          ;chk new offset
minoff  ADDI.L #modulo,D0
        BMI.B minoff
pluoff  CMPI.L #modulo,D0
        BCS.B newoff
        SUBI.L #modulo,D0
        BRA.B pluoff
newoff  MOVE.L D0,#$000A(A1)  ;save new offset
        SWAP D0               ;chk new CTR offset
        TST.W,D0
        BNE.B ctroff
        MOVEQ #1,D0
ctroff  MOVE.W D0,D1
        MOVE.B D0,(#ic1ctr0)  ;offset to ic1ctr0
        ASR.W #8,D0
        MOVE.B D0,(#ic1ctr0)
        MULU #$0010(A1),D1    ;ctr offset to PLL phase offset
        MOVE.W D1,(A1)+
        SUBQ.L #8,D3          ;inc carrier phase
        MOVE.W (A1)+,D0
        MOVE.W (A1),D1
        MULS D3,D1
        ADD.W D1,D0
        MOVE.W D0,-(A1)
        MOVEQ #6,D0           ;set D0 back to 6

;set channel area address to A0

chadd   MOVEA.L A0,A1         ;A0 to A1, set channel #
        MOVE.W (A0)+,D1
        MOVE.W (A0)+,D2
        MOVE.W D1,D3
        ANDI.W #1,D3
        BNE.B datach
        EXT.L D1              ;not data channel
        DIVU #3,D1
        EXT.L D1
        DIVU #3,D1
        SWAP D1
        ADDQ.W #1,D1
        ADD.W D1,D2
datach  ADDQ.L #8,A0          ;set ch adr A0
        ANDI.L #3,D2
        LSL.L #8,D2
        LSL.L #1,D2
        ADDA.L D2,A0

;count 7 - set next channel code lookup table & RF channel PLL

        BSR.W srx
        CMPI.W #7,D0          ;count 7?
        BCS.W cstx
        MOVE.W (A1),D1        ;rotate channel sequence
        ADDQ.W #1,D1
        CMPI.W #18,D1
        BCS.B clrd1
        CLR.W D1
clrd1   MOVE.W D1,(A1)
        MOVE.W #$0002(A1),D2  ;set channel #
        MOVE.W D1,D3
        ANDI.W #1,D3
        BNE.B datchn
        EXT.L D1              ;not data channel
        DIVU #3,D1
        EXT.L D1
        DIVU #3,D1
        SWAP D1
        ADDQ.W #1,D1
        ADD.W D1,D2
datchn  ANDI.W #3,D2
        ASL.W #6,D2
        MOVE.W D2,D1
        ADD.B D2,D2           ;address bit 1
        BCS.B adbit1s
        CLR.B (#rbit1)
        BRA.B adbit1r
adbit1s CLR.B (#sbit1)
adbit1r ADD.B D2,D2           ;address bit 0
        BCS.B adbit0s
        CLR.B (#rbit0)
        BRA.B rfchn
adbit0s CLR.B (#sbit0)
rfchn   ASL.W #3,D1           ;set RF channel - modulo to PLL
        MOVEQ #0,D0
        MOVE.W #$001E(A2,D1.W),D0
        ADDQ.L #4,D0
        DIVU #10,D0
        NEG.W D0
        ANDI.W #15,D0
        MOVE.W D0,D1
        SWAP D0
        ADDQ.W #6,D0
        NEG.W D0
        ANDI.W #15,D0
        ASL.W #4,D0
        ADD.W D1,D0
        MOVEQ #7,D1           ; loop on 8 bits
rfchn3  ADD.B D0,D0           ; set data bit
        BCC.B rfchn1
        CLR.B (#sdbit)
        BRA.B rfchn2
rfchn1  CLR.B (#rdbit)
rfchn2  CLR.B (#scbit)        ; clock pulse
        CLR.B (#rcbit)
        DBF D1,rfchn3
        CLR.B (#ssbit)        ; strobe pulse
        CLR.B (#rsbit)
        CLR.B (#rdbit)        ; reset data bit
        MOVEQ #7,D0           ; set D0 back to 7
cstx    BSR.W stx

;carrier & PSK data demodulator Costas-loop PLL

        MOVE.W (A0)+,D3       ;inc carrier phase
        MOVE.W (A0)+,D1
        MOVE.W (A0)+,D2
        ADD.W D2,D1
        MOVE.W D1,#$FFFC(A0)
        SUB.W D3,D1           ;sine table offset @ $1040
        LSR.W #8,D1
        LSL.W #1,D1
        ADDI.W #$1080,D1
        ADDQ.L #2,A0
        MOVE.L D4,D6
        MOVE.L D5,D7
        MULS #$C0(A2,D1.W),D4 ;rotated I
        MULS #$40(A2,D1.W),D7
        SUB.L D7,D4
        ADD.L D4,D4
        SWAP D4
        MULS #$40(A2,D1.W),D6 ;rotated Q
        MULS #$C0(A2,D1.W),D5
        ADD.L D6,D5
        ADD.L D5,D5
        SWAP D5
        MOVE.W D0,D1          ;move data to 8-sample buffer
        ADD.W D1,D1
        MOVE.W D4,#$0E(A0,D1.W)
        MOVE.W D4,D1          ;loop AGC
        BPL.B abs5
        NEG.W D1
abs5    MOVE.W D5,D2
        BPL.B abs6
        NEG.W D2
abs6    ADD.W D2,D1
        ADDQ.W #1,D1
        CMPI.W #$0400,D1
        BCS.B agcoff
        SWAP D4
        ASR.L #6,D4
        DIVS D1,D4
        SWAP D5
        ASR.L #6,D5
        DIVS D1,D5
agcoff  MOVE.W D5,D2          ;lock detector D2.W
        BPL.B abs7
        NEG.W D2
abs7    MULS D5,D4            ;I*Q to D4,D5
        MOVE.L D4,D5
        ASR.L #6,D4           ;correct carrier phase
        ADD.W D4,#$FFFA(A0)
        ASL.L #4,D5           ;correct carrier frequency
        ADD.L D5,#$FFFC(A0)
        MOVE.W (A0),D7        ;lock?
        CMPI.W #$00A0,D2
        BCS.B lockinc
        SUBQ.W #1,D7
        BRA.B lockdec
lockinc ADDQ.W #1,D7
lockdec TST.W D7
        BPL.B locknz
        MOVEQ #0,D7
locknz  CMPI.W #maxcar,D7
        BCS.B lockov
        MOVE.W #maxcar,D7
lockov  MOVE.W D7,(A0)        ;force carrier search?
        BNE.B codepll
        ADDI.W #$0080,#$FFFC(A0)

;end of accumulation, compute & check code-DCD

codepll SUBQ.W #7,D0          ;end of accumulation?
        BPL.B endofa
        RTS
endofa  CLR.W #$0016(A2)
        MOVE.L #$000E(A2),D0  ;average /16
        ASR.L #4,D0
        MOVE.L #$0012(A2),D1
        ASR.L #4,D1
        MOVE.L D0,D7          ;code DCD treshold 512
        SUBI.W #512,D7
        ADDA.W #$000A,A0      ;new code DCD
        MOVEQ #1,D3
        MOVE.W (A0),D4
        TST.W D7
        BPL.B dcdinc
        MOVEQ #-1,D3
dcdinc  ADD.W D3,D4
        BPL.B dcdnz           ;chk code DCD
        MOVEQ #0,D4
dcdnz   CMPI.W #maxcod,D4
        BCS.B dcdov
        MOVE.W #maxcod,D4
dcdov   MOVE.W D4,(A0)+
        BNE.B dcdok           ;invalid code dcd
        MOVE.L #$00090000,D3
        MOVE.W (A0),#$FFFA(A0)
        CLR.W #$FFF4(A0)
        BRA.B newcof
dcdok   MOVEQ #0,D3           ;valid data?
        TST.W D7
        BMI.B invdat
        MOVE.W D1,D3          ;compute correction
        NEG.W D3
        SWAP D3
        CLR.W D3
        ASR.L #2,D3
        DIVS D0,D3
        EXT.L D3
invdat  MOVE.L D3,D4          ;rate correction
        ASL.L #7,D4
        SWAP D4
        ADD.W D4,#$FFFA(A0)
newcof  MOVEA.L #$FFF6(A0),A1 ;offset increment/correction
        ADD.L D3,#$FFF6(A0)
        ADDQ.L #2,A0

;check PLL frequency, find transition

        MOVEQ #0,D0
        MOVE.W (A0)+,D0
        MOVE.L D0,D1
        MOVEQ #6,D3
        MOVEQ #0,D4
        MOVEQ #0,D5
        MOVE.L #$CCCCCCCC,D6
trloop  MOVE.W (A0),D2        ;transition found?
        EOR.W D0,D2
        BPL.B notran
        MOVE.W D0,D6          ;compute fractional part
        BPL.B abs8
        NEG.W D6
abs8    MOVE.W (A0),D2
        BPL.B abs9
        NEG.W D2
abs9    ADD.W D2,D6
        SWAP D2
        MOVE.W #$FFFF,D2
        DIVU D6,D2
        MOVE.W D3,D6
        SWAP D6
        MOVE.W D2,D6
notran  MOVE.W (A0)+,D2       ;check carrier frequency
        EXT.L D2
        ADD.L D2,D0
        BPL.B abs10
        NEG.L D0
abs10   ADD.L D0,D4
        SUB.L D2,D1
        BPL.B abs11
        NEG.L D1
abs11   ADD.L D1,D5
        MOVE.L D2,D0
        MOVE.L D2,D1
        DBF D3,trloop
        TST.W D7              ;frequency error?
        BMI.B bitsyn
        SUB.L D5,D4
        BCC.B bitsyn
        ADDI.W #$2000,#$FFDE(A0)

;100bps half-bit synchronization

bitsyn  MOVE.L A1,D2          ;add code offset
        DIVU #moddiv,D2
        SWAP D2
        CLR.W D2
        SWAP D2
        ADD.L D2,D6
        SWAP D6
        EXT.L D6
        MOVE.L A1,D2          ;old/new offset difference
        SWAP D2
        MOVE.W (A0),D3
        MOVE.W D2,(A0)+
        SUB.W D3,D2
        CMPI.W #256,D2        ;correct count
        BMI.B bsyn1
        SUBQ.L #1,(A0)
bsyn1   ADDI.W #256,D2
        BPL.B bsyn2
        ADDQ.L #1,(A0)
bsyn2   MOVE.L #$FFD4(A0),D0  ;current 1ms, reference 1ms
        MOVE.L (A0),D1
        MOVEA.L D1,A1
        MOVE.L D0,D3          ;update reference
        SUB.L D1,D3
        DIVU #10,D3
        MULU #10,D3
        ADD.L D3,D1
        MOVE.L D1,(A0)+
        MOVE.W (A0),D2        ;DCD?, no transition?
        TST.W D7
        BMI.B nobitt
        TST.W D6
        BMI.B nobitt
        MOVE.L D0,D3          ;first transition?
        SUB.L D6,D3
        TST.W D2
        BNE.B bitdcd
        MOVE.L D3,#$FFFC(A0)
        MOVEQ #1,D2
        BRA.B bitchk
bitdcd  SUB.L D1,D3           ;modify bit-dcd
        BNE.B bitdec
        ADDQ.W #1,D2
        BRA.B bitchk
bitdec  SUBQ.W #1,D2
bitchk  TST.W D2              ;check bit-dcd
        BPL.B bitov
        MOVEQ #0,D2
        BRA.B nobitt
bitov   CMPI.W #maxbit,D2
        BCS.B nobitt
        MOVE.W #maxbit,D2
nobitt  TST.W #$FFE6(A0)      ;code-dcd resets bit-dcd
        BNE.B nobitr
        MOVEQ #0,D2
nobitr  MOVE.W D2,(A0)+

;accumulate one bit data
        SUBQ.L #7,D0          ;D0.L=bit-ref, D1.L=new-ref
        MOVE.L (A0),D3        ;D3.L=half-bit-accu
        MOVE.L D1,D6          ;two half-bits?
        CMP.L D0,D1
        BCS.B bcompl
        SUBI.L #10,D6
bcompl  BSR.W databit
        MOVEQ #0,D3           ;dump half-bit
        MOVE.L D1,D6          ;D6.L=half-bit-ref
        MOVEQ #7,D2
        MOVEQ #0,D4           ;D4.W=sample_offset
accloop MOVE.W #$00E8(A0,D4.W),D5     ;get data sample
        ADDQ.W #2,D4
        EXT.L D5
        CMP.L D1,D0           ;acc. or dump half-bit?
        BEQ.B dump
        ADD.L D5,D3
        BRA.B nodump
dump    BSR.W databit
        MOVEQ #0,D3
nodump  ADDQ.L #1,D0
        DBF D2,accloop
        MOVE.L D3,(A0)+       ;save half-bit accumulator

;average results to buffer

        ASR.W #4,D7 ;signal level - treshold average
        ADD.W D7,(A0)+
        MOVE.L #$FFD0(A0),D0  ;carrier frequency average
        ASR.L #4,D0
        ADD.L D0,(A0)+
        MOVE.L #$FFD2(A0),D0  ;code offset average
        SWAP D0
        TST.W #$0008(A0)
        BNE.B fstavg
        MOVE.W D0,(A0)
fstavg  SUB.W (A0),D0
        CMPI.W #-999,D0
        BPL.B fstavg1
        ADDI.W #moddiv,D0
fstavg1 CMPI.W #999,D0
        BMI.B fstavg2
        SUBI.W #moddiv,D0
fstavg2 ADD.W (A0)+,D0
        SWAP D0
        ASR.L #4,D0
        ADD.L D0,(A0)+
        MOVE.W #$FFC0(A0),D0  ;epoch average *16
        ADD.W D0,(A0)+
        ADDQ.W #1,(A0)        ;averaging /16
        CMPI.W #16,(A0)
        BCC endavg
        RTS
endavg  CLR.W (A0)+        ;end of averaging, results to buffer
        MOVE.W (A0),D0     ;buffer pointer
        ADDI.W #16,D0
        ANDI.W #240,D0
        MOVE.W D0,(A0)+
        MOVE.L #$FFB8(A0),#0(A0,D0.W)  ;end 1ms epoch
        MOVE.L #$FFF0(A0),#4(A0,D0.W)  ;carrier freq. average
        CLR.L #$FFF0(A0)
        MOVE.L #$FFF6(A0),D1           ;code offset average
        BPL.B endavg1
        ADDI.L #modulo,D1
endavg1 CMPI.L #modulo,D1
        BMI.B endavg2
        SUBI.L #modulo,D1
endavg2 MOVE.L D1,#8(A0,D0.W)
        CLR.L #$FFF6(A0)
        MOVE.W #$FFEE(A0),#12(A0,D0.W) ;signal level average
        CLR.W #$FFEE(A0)
        MOVE.W #$FFFA(A0),#14(A0,D0.W)  ;.W epoch average *16
        CLR.W #$FFFA(A0)
        RTS

;data bits to RX data buffer

databit MOVEM.L #$F8E0,-(A7)  ;save regs. on stack
        BSR.B procpri
        MOVEM.L (A7)+,#$071F
        RTS
procpri MOVE SR,D0            ;set proc. priority 3
        ANDI.W #$F8FF,D0
        ORI.W #$0300,D0
        MOVE D0,SR
        ADDA.W #$0126,A0      ;A0 to offset $0178

        MOVE.L (A0),D1        ;manchester decoder
        MOVE.L D3,(A0)+
        SUB.L D1,D3
        SWAP D3
        MOVE.L (A0),D1        ;differential decoding
        SWAP D1
        MOVE.W D1,D2
        MOVE.W D3,D1
        MOVE.L D1,(A0)+
        EOR.W D2,D3
        MOVEQ #10,D0          ;shift last 176 bits
        ADD.W D3,D3
loop176 ROXL.W (A0)+
        DBF D0,loop176
        ADDQ.W #1,(A0)        ;counter /200
        CMPI.W #200,(A0)
        BCC.B hamming
        RTS

;(85,77) Hamming code coefficients
hamming LEA.L #4(PC),A1       ;lookup table start address to A1
        BRA.B parity
        $00000000D3D2D1D0                   ;88-81, 88-85 dummy
        $CFCECDCCCBCAC9C8C7C6C5C4C3C2C1BF   ;80-65
        $BEBDBCBBBAB9B8B7B6B5B4B3B2B1B0AF   ;64-49
        $AEADACABAAA9A8A7A6A5A4A3A2A19F9E   ;48-33
        $9D9C9B9A9998979695949392918F8E8D   ;32-17
        $8C8B8A898786858380C0A09088848281   ;16-1

parity  MOVEA.L A0,A2         ;compute parity to D0.B
        MOVEQ #0,D0
        MOVEQ #10,D1
wploop  MOVE.W -(A2),D2
        MOVEQ #7,D3
bploop  MOVE.B (A1)+,D4
        ADD.W D2,D2
        ADD.W D2,D2
        BCC.B bitclr
        EOR.B D4,D0
bitclr  DBF D3,bploop
        DBF D1,wploop
        TST.B D0              ;parity check ok?
        BEQ.B parok
        RTS
parok   MOVEA.L A0,A2         ;reset /200 ctr
        MOVE.W (A0),D5
        CLR.W (A0)+

;verified data to data buffer
        MOVE.W (A0),D0        ;data buffer pointer
        ADDI.W #16,D0
        ANDI.W #112,D0
        MOVE.W D0,(A0)+
        MOVE.W D5,#0(A0,D0.W) ;/200 ctr to buffer
        ADDQ.W #2,D0
        MOVEQ #4,D5           ;shift data to buffer
ddloop  MOVEQ #0,D4
        MOVEQ #1,D1
wdloop  MOVE.W -(A2),D2
        MOVEQ #7,D3
bdloop  ADD.W D2,D2
        ADD.W D2,D2
        ROXL.W #1,D4
        DBF D3,bdloop
        DBF D1,wdloop
        MOVE.W D4,#0(A0,D0.W)
        ADDQ.W #2,D0
        DBF D5,ddloop
        MOVE.L D6,#0(A0,D0.W) ;1ms ctr to buffer
        RTS

;RS-232 input variables #offset(A2)
;$1410 256*byte: 256 character buffer space
;$1510 word: RX timer
;$1512 word: RX shift register
;$1514 byte: write buffer pointer
;$1515 byte: read buffer pointer

portc   EQU $00010005
inbit   EQU 6

;RS-232 serial receive routine (100bps)

srx     MOVE.W #$1510(A2),D1        ;RX timer to D1
        ADDQ.W #1,D1                ;increment RX timer
        CMPI.W #$0008,D1            ;start bit?
        BCC.B srx1
        BTST #inbit,(#portc)
        BEQ.B srx2
        CLR.W D1
srx2    MOVE.W D1,#$1510(A2)
        RTS
srx1    CMPI.W #$0905,D1            ;stop bit?
        BCS.B srx3
        BTST #inbit,(#portc)
        BEQ.B srx4
        MOVE.W #$1400,D1            ;RX character to buffer
        MOVE.B #$1514(A2),D1
        MOVE.B #$1513(A2),#$10(A2,D1.W)
        ADDQ.B #1,D1
        MOVE.B D1,#$1514(A2)
srx4    CLR.W #$1510(A2)
        RTS
srx3    CMPI.B #$0A,D1              ;data bit counter
        BCS.B srx5
        ADDI.W #$00F6,D1
srx5    MOVE.W D1,#$1510(A2)
        CMPI.B #$05,D1
        BNE.B srx6
        LSR.W #$1512(A2)            ;RX shift register
        BTST #inbit,(#portc)
        BEQ.B srx6
        ADDI.W #$0080,#$1512(A2)    ;"1" to shift register
srx6    RTS

;RS-232 output variables #offset(A2)
;$1300 256*byte: 256 character buffer space
;$1400 word: transmitted character buffer
;$1402 byte: write buffer pointer
;$1403 byte: read buffer pointer

;RS-232 serial transmit routine (1000bps)

stx     LSR.W #$1400(A2)            ;shift serial register
        BEQ.B stx1
        BCC.B stx2                  ;serial bit output
        CLR.B (#rdbit)              ;transmit 1
        BRA.B stx3
stx2    CLR.B (#sdbit)              ;transmit 0
        BRA.B stx3
stx1    CLR.B (#rdbit)              ;stop bits
        MOVE.W #$1300,D1
        MOVE.B #$1403(A2),D1        ;buffer empty?
        CMP.B #$1402(A2),D1
        BEQ.B stx3
        MOVE.W #$0300,#$1400(A2)    ;new data to shift reg.
        MOVE.B #$0(A2,D1.W),#$1401(A2)
        ADDQ.B #1,#$1403(A2)
        CLR.B (#sdbit)              ;start bit
stx3    RTS

;*** end of INT4 routine ***
